#include "Composite_interaction.h"

namespace MM
{

double Composite_interaction::
potential ()
{
    double result = 0;

    if (common_.was_initialized() && common_on_)
        result += common_().potential ();

    if (addition_on_)
        for (int i=0;  i<interaction_.size();  ++i)
            result += interaction_[i].potential ();

    return result;
}

void Composite_interaction::
add_force ()
{
    if (common_.was_initialized() && common_on_)
        common_().add_force ();

    if (addition_on_)
        for (int i=0;  i<interaction_.size();  ++i)
            interaction_[i].add_force ();
}

void Composite_interaction::
add_long_force ()
{
    if (common_.was_initialized() && common_on_)
        common_().add_long_force ();

    if (addition_on_)
        for (int i=0;  i<interaction_.size();  ++i)
            interaction_[i].add_long_force ();
}

void Composite_interaction::
add_short_force ()
{
    if (common_.was_initialized() && common_on_)
        common_().add_short_force ();

    if (addition_on_)
        for (int i=0;  i<interaction_.size();  ++i)
            interaction_[i].add_short_force ();
}

void Composite_interaction::
report (Text *result, Text const & hint)
{
    if (addition_on_)
        for (int i=0;  i<interaction_.size();  ++i)
            interaction_[i].report (result, hint);

    if (common_.was_initialized() && common_on_)
        common_().report (result, hint);
}

void Composite_interaction::
adopt (Interaction *adoptee)
{
    interaction_.adopt (adoptee);
}

void Composite_interaction::
clear ()
{
    interaction_.clear ();
}

bool Composite_interaction::
locked () const
{
    bool result = true;

    for (int i=0;  i<interaction_.size();  ++i)
        result &= interaction_[i].locked ();

    if (common_.was_initialized())
        result &= common_().locked ();

    return result;
}

void Composite_interaction::
set_lock (bool on)
{
    for (int i=0;  i<interaction_.size();  ++i)
        interaction_[i].set_lock (on);

    if (common_.was_initialized())
        common_().set_lock (on);
}

}//MM
