#include "List_interactions.h"

#include "Log.h"

#include "Create.h"
#include "Timer.h"
#include "Model_kit.h"
#include "anint.h"
#include "Profiler.h"

#include <math.h>

namespace MM
{

//List_interactions::
//List_interactions (Atom_group & atom_group, Force_field & ff)
//:
//    Yakub_Ronchi_interactions (atom_group, ff),
//    pair_list_(0), is_locked_(false)
//{
//}
//
//List_interactions::
//List_interactions (Atom_group & atom_group,
//    Bond_group & bond_group, Force_field & ff)
//:
//    Yakub_Ronchi_interactions (atom_group, bond_group, ff),
//    pair_list_(0), is_locked_(false)
//{
//}

List_interactions::
List_interactions (Model & model/*, Force_field & ff*/)
:
    Yakub_Ronchi_interactions (model/*, ff*/),
    pair_list_(0), is_locked_(false)
{
}


void List_interactions::
set_lock (bool on)
{
    //Bond_group_cache::set_lock (on);

    if (on)
    {
        make_list ();
        is_locked_ = true;
    }
    else
    {
        is_locked_ = false;
    }
}


bool List_interactions::locked () const
{
    return is_locked_;
}

void List_interactions::
make_list ()
{
    if (!cache_flag_)
    {
        cache_atoms ();
        cache_bonds ();
    }

    make_volumes ();//fix

    //int n = pair_list_.size();
    //pair_list_.clear();
    //pair_list_.reserve (n);

    int i, j, count = atom_count();

    delete [] pair_list_;
    pair_list_ = new Next_atom [count*count];

    //for(j=0;  j<count;  ++j)
    //{
    //    Atom_cache & atom_j = atom_[j];

    //    miss_neighbours (atom_j);

    //    Force_field::VdWaals * vdw_ptr = 
    //        force_field_.VdW_par [atom_j.mm_type___];
    //    
    //    for (i=j+1;  i<count;  ++i)
    //    {
    //        Atom_cache & atom_i = atom_[i];
    //        
    //        if (atom_i.transparency___ == 0.)
    //            continue;

    //        Force_field::VdWaals & vdw = vdw_ptr [atom_i.mm_type___];
    //        pair_list_.push_back (Atom_pair (j, i, vdw.A, vdw.B));
    //    }
    //    
    //    process_14 (atom_j, None);
    //    restore_neighbours ();
    //}

    Force_field & ff = force_field();

    for(j=0;  j<count;  ++j)
    {
        Atom_cache & atom_j = atom_[j];

        miss_neighbours (atom_j);

        Force_field::VdWaals * vdw_ptr = 
            ff.VdW_par [atom_j.mm_type___];
        
        for (i=0;  i<count;  ++i)
        {
            Atom_cache & atom_i = atom_[i];
            
            double miss = 
                atom_i.transparency___ * atom_j.transparency___ * (i != j);

            Force_field::VdWaals & vdw = vdw_ptr [atom_i.mm_type___];
            //pair_list_.push_back (Atom_pair (/*j, */i, 
            //    (float)vdw.A, (float)vdw.B, (float)miss));
            pair_list_[j*count + i] = Next_atom (/*j, */i, 
                (float)vdw.A, (float)vdw.B, (float)miss);
        }
        
        process_14 (atom_j, No);
        restore_neighbours ();
    }

    if (!cache_flag_)
    {
        atom_.clear ();
        bond_.clear ();
    }
}

void List_interactions::
cache_coords ()
{
    REQUIRE ("Atoms cached.", locked());

    int i, count = atom_group_.atom_count();
    
    for (i=0;  i<count;  ++i)
    {
        Atom & current = atom_group_.atom (i);
        Atom_cache & cache = atom_[i];
        cache.x___ = current.x();
        cache.y___ = current.y();
        cache.z___ = current.z();
    }
}


double List_interactions::
potential ()
{
    int i, count;
    last_potential_  = 0.;

    Van_der_Waals_potential_    = 0.;
    electrostatic_potential_    = 0.;
    bond_potential_             = 0.;
    angle_potential_            = 0.;
    torsion_potential_          = 0.;
    improper_torsion_potential_ = 0.;
    hydrophobic_potential_      = 0.;
    surface_area_potential_     = 0.;
    generalized_Born_potential_ = 0.;

    if (!cache_flag_)
    {
        cache_atoms ();
        cache_bonds ();

        if (!locked())
        {
            cache_flag_ = true; //fix !!! always
            make_list   ();
            cache_flag_ = false;
        }
    }

    try
    {
        full_nonbonded_potential ();

        count = atom_count();
        for (i=0;  i<count;  ++i)
        {
            Atom_cache & current_atom = atom_[i];
            atom_potential (current_atom);
        }

        if (&bond_group_ != &nil<Bond_group>())
        {
            count = bond_count();
            for (i=0;  i<count;  ++i)
            {
                Bond_cache & current_bond = bond_[i];             
                bond_potential (current_bond);
            }
        }
    }
    catch(...)
    {
        to_user().fatal_error ("List_interactions: Unknown exception.");
    }

    if (!cache_flag_)
    {
        //flush_born  ();//fix down
        atom_.clear ();
        bond_.clear ();
    }

    return last_potential_;
}

void List_interactions::
add_force ()
{
    Timer timer;
    Text  time;

    int i, count;

    profile().List_interactions___add_force___start ();

    if (!cache_flag_)
    {
        cache_atoms ();
        cache_bonds ();

        if (!locked())
        {
            cache_flag_ = true; //fix !!! always
            make_list   ();
            cache_flag_ = false;
        }
    }

    profile().List_interactions___add_force___make_list ();

    time += "   cache ";
    time += timer.elapsed();    timer.restart();

    try
    {
        zero_forces ();      //fix remove?

        calc_nonbonded_force ();

        profile().List_interactions___add_force___calc_nonbonded_force ();


        time += "   nb ";
        time += timer.elapsed();    timer.restart();

        count = atom_count ();
        for (i=0;  i<count;  ++i)
            add_atom_force (atom_[i]);

        profile().List_interactions___add_force___add_atom_force ();

        time += "   atom ";
        time += timer.elapsed();    timer.restart();

        if (&bond_group_ != &nil<Bond_group>())
        {
            count = bond_count ();

            for (i=0;  i<count;  ++i)
                add_bond_force (bond_[i]);
        }

        profile().List_interactions___add_force___add_bond_force ();

        time += "   bond ";
        time += timer.elapsed();    timer.restart();

        add_full_nonbonded_force ();

        profile().List_interactions___add_force___add_full_nonbonded_force ();

        time += "   add nb ";
        time += timer.elapsed();    timer.restart();
    }
    catch(...)
    {
        to_user().fatal_error ("List_interactions: Unknown exception.");
    }

    if (!cache_flag_)
    {
        flush_force ();
        //flush_born  ();//fix down
        atom_.clear ();
        bond_.clear ();
    }

    profile().List_interactions___add_force___flush_force ();

    time += "   flush ";
    time += timer.elapsed();    timer.restart();
    //to_user().status (time);
    to_user().clock ();
}

void List_interactions::
full_nonbonded_potential ()
{
    Timer timer;

    double Rm      = force_field().Rm_;

    double  L = 0., lIL = 0., LI2 = 0.;
    bool    box = false;
    if (model_.kit().boundary_conditions ().type_name() == "Box")
    {
        box = true;
        L = model_.kit().boundary_conditions ().box ().a_mod_rectangle();//fix
        lIL = 1. / L;
        LI2 = L / 2.;

        if (Rm > LI2)
            Rm = LI2;
    }

    const double Rm_2    = Rm * Rm;
    const double lIRm    = 1. / Rm;

    double  lIepsilon = Constant::lI4PiEps0 / epsilon();
    double  A = 0., B = 0.;
    double  x, y, z;
    double  r_2, lIr, lIr_6, lIr_eps;
    double  P_VdW = 0., P_el = 0., Yakub_Ronchi = 0.;

    //int count = pair_list_.size();

    //for (int i=0;  i<count;  ++i)
    //{
    //    Atom_pair const & pair = pair_list_[i];
    //    Atom_cache & atom1 = atom_[pair.atom1_N___];
    //    Atom_cache & atom2 = atom_[pair.atom2_N___];

    //    A = pair.A___;
    //    B = pair.B___;

    //    x = atom1.x___ - atom2.x___;
    //    y = atom1.y___ - atom2.y___;
    //    z = atom1.z___ - atom2.z___;

    //    r_2     = x*x + y*y + z*z;
    //    lIr     = 1.0 / sqrt (r_2);
    //    lIr_eps = lIr * lIepsilon;
    //    lIr_6   = lIr * lIr;    lIr_6 *= lIr_6 * lIr_6;

    //    P_VdW += (A * lIr_6 - B ) * lIr_6;
    //    P_el  += atom1.charge___ * atom2.charge___ * lIr_eps; //fix
    //}
    int count = atom_count();

    for (int j=0;  j<count;  ++j)
    {
        Atom_cache & atom1 = atom_[j];

        double  ax = atom1.x___;
        double  ay = atom1.y___;
        double  az = atom1.z___;

        int count_i = j*count + count;

        for (int i=j*count + j +1 ;  i<count_i;  ++i)
        {
            Next_atom const & pair = pair_list_[i];
            Atom_cache & atom2 = atom_[pair.atom2_N___];

            A = pair.A___;
            B = pair.B___;

            x = ax - atom2.x___;
            y = ay - atom2.y___;
            z = az - atom2.z___;

            if (box)
            {
                x -= L * anint (x * lIL);
                y -= L * anint (y * lIL);
                z -= L * anint (z * lIL);
            }

            r_2     = x*x + y*y + z*z + 0.000001;
            lIr     = pair.miss___ * 1.0 / sqrt (r_2);
            lIr_eps = lIr * lIepsilon;
            lIr_6   = lIr * lIr;    lIr_6 *= lIr_6 * lIr_6;

            double rIRm = r_2 * lIr * lIRm;
            Yakub_Ronchi   = 1. + .5 * rIRm * (rIRm * rIRm - 3.); 

            P_VdW += (A * lIr_6 - B ) * lIr_6;

            P_el  += (r_2 < Rm_2) * Yakub_Ronchi *
                atom1.charge___ * atom2.charge___ * lIr_eps; //fix
        }
    }

    Van_der_Waals_potential_ += P_VdW;
    electrostatic_potential_ += P_el;
    //hydrophobic_potential_   += P_hph;

    last_potential_ += P_el + P_VdW /*+ P_hph*/;

    //log() << "\n Nonbonded " << timer.elapsed();
}

void List_interactions::
add_full_nonbonded_force ()
{
    double Rm      = force_field().Rm_;

    double  L = 0., lIL = 0., LI2 = 0.; 
    bool    box = false;
    if (model_.kit().boundary_conditions ().type_name() == "Box")
    {
        box = true;
        L = model_.kit().boundary_conditions ().box ().a_mod_rectangle();//fix
        lIL = 1. / L;
        LI2 = L / 2.;

        if (Rm > LI2)
            Rm = LI2;
    }

    const double Rm_2    = Rm * Rm;
    const double lIRm_3  = 1. / (Rm*Rm*Rm);

    //double  lIepsilon = EPS0 / epsilon();
    double  lIepsilon = Constant::lI4PiEps0 / epsilon();

    //////int count = pair_list_.size();

    //for (int i=0;  i<count;  ++i)
    //{
    //    Atom_pair const & pair = pair_list_[i];
    //    Atom_cache & atom1 = atom_[pair.atom1_N___];
    //    Atom_cache & atom2 = atom_[pair.atom2_N___];

    //    double x = atom1.x___ - atom2.x___;
    //    double y = atom1.y___ - atom2.y___;
    //    double z = atom1.z___ - atom2.z___;

    //    double A = pair.A___;
    //    double B = pair.B___;

    //    double r_2     = x*x + y*y + z*z;
    //    double lIr     = 1.0 / sqrt (r_2);
    //    double lIr_eps = lIr * lIepsilon;
    //    double lIr_2   = lIr * lIr;    
    //    double lIr_6   = lIr_2 * lIr_2 * lIr_2;

    //    double q = atom1.charge___ * atom2.charge___ * lIr_eps;
    //    double K = - (6 * (B - 2. * A * lIr_6) * lIr_6 - q) * lIr_2;

    //    double fx = K * x;
    //    double fy = K * y;
    //    double fz = K * z;

    //    atom1.force_x___ += fx;
    //    atom2.force_x___ -= fx;

    //    atom1.force_y___ += fy;
    //    atom2.force_y___ -= fy;

    //    atom1.force_z___ += fz;
    //    atom2.force_z___ -= fz;
    //}

    int count = atom_count();

    for (int j=0;  j<count;  ++j)
    {
        Atom_cache & atom1 = atom_[j];

        double  ax = atom1.x___;
        double  ay = atom1.y___;
        double  az = atom1.z___;

        int count_i = j*count + count;

        for (int i=j*count + j +1 ;  i<count_i;  ++i)
        {
            Next_atom const & pair = pair_list_[i];
            Atom_cache & atom2 = atom_[pair.atom2_N___];

            double x = ax - atom2.x___;
            double y = ay - atom2.y___;
            double z = az - atom2.z___;

            if (box)
            { 
                //x = fmod (x, LI2);
                //y = fmod (y, LI2);
                //z = fmod (z, LI2);

                //if      (x >  LI2) x -= L;
                //else if (x < -LI2) x += L;
                //if      (y >  LI2) y -= L;
                //else if (y < -LI2) y += L;
                //if      (z >  LI2) z -= L;
                //else if (z < -LI2) z += L;

                x -= L * anint (x * lIL);
                y -= L * anint (y * lIL);
                z -= L * anint (z * lIL);
            }

            double A = pair.A___;
            double B = pair.B___;

            double r_2     = x*x + y*y + z*z + 0.000001;
            double lIr     = pair.miss___ * (1.0 / sqrt (r_2));
            double lIr_2   = lIr * lIr;    
            double lIr_6   = lIr_2 * lIr_2 * lIr_2;

            double Kf_Yakub_Ronchi = pair.miss___ * lIRm_3 - lIr_2 * lIr; 

            double q = - atom1.charge___ * atom2.charge___ * lIepsilon
                * Kf_Yakub_Ronchi * (r_2 < Rm_2);

            double VdW = - (6 * (B - 2. * A * lIr_6) * lIr_6) * lIr_2;
            double K = q + VdW;

            double fx = K * x;
            double fy = K * y;
            double fz = K * z;

            atom1.force_x___ += fx;
            atom2.force_x___ -= fx;

            atom1.force_y___ += fy;
            atom2.force_y___ -= fy;

            atom1.force_z___ += fz;
            atom2.force_z___ -= fz;
        }
    }
}

void List_interactions::
atom_potential (Atom_cache & atom)
{
    miss_neighbours (atom);
    process_14      (atom, Potential);//fix ???
    restore_neighbours ();


    angle_potential            (atom);
    improper_torsion_potential (atom);
}

void List_interactions::
add_atom_force (Atom_cache & atom)
{
    miss_neighbours (atom);
    process_14      (atom, Force);//fix ???
    restore_neighbours ();

    add_angle_force            (atom);
    add_improper_torsion_force (atom);
}

}//MM

