#include "Still_interactions.h"

#include "Log.h"

#include "Defs.h"
#include "Mach_eps.h"
#include "Timer.h"

#include <math.h>

namespace MM
{
//double Still_interactions:: Sigma =  0.0049;
double Still_interactions:: Sigma =  0.009;

double Still_interactions:: Shift   = 0.000001;
//double Still_interactions:: Shift   = pow (2., 2);


double Still_interactions:: Mbond   = 1.01;

double Still_interactions:: RH      = 1.15;
//double Still_interactions:: RH      = 1.15 * 1.05;

double Still_interactions:: R14      = 3.;

//double Still_interactions:: Sigma =  0.0081;    // Shift = 1.2
//double Still_interactions:: Sigma =  0.0081;    // Shift = 2

//double Still_interactions:: Sigma =  0.0049;
//double Still_interactions:: Sigma =  0.0087;    // Shift = 5
//double Still_interactions:: Sigma =  0.0085;    // Shift = 4
//double Still_interactions:: Sigma =  0.0082;    // Shift = 3
//double Still_interactions:: Sigma =  0.0080;    // Shift = 2
//double Still_interactions:: Sigma =  0.0075;    // Shift = 1.7

double Still_interactions:: Phi   = -0.09;

double Still_interactions:: R_probe = 1.4;

//double Still_interactions:: Shift   = pow (1.7, 2);
//double Still_interactions:: Shift   = pow (3., 2);

//double Still_interactions:: Power = 6;
double Still_interactions:: Power = 4;
//double Still_interactions:: Power = 3;
//double Still_interactions:: Power = 2;

//double Still_interactions:: P1 =  0.073;
double Still_interactions:: P1 =  0.0;
//double Still_interactions:: P1 =  0.273;
//double Still_interactions:: P1 =  0.15;
//double Still_interactions:: P1 =  0.6;

//double Still_interactions:: P2 =  0.921;
//double Still_interactions:: P3 =  6.211;
double Still_interactions:: P2 =  0.47;
double Still_interactions:: P3 =  5.76;
//double Still_interactions:: P2 =  0.;
//double Still_interactions:: P3 =  0.;

//double Still_interactions:: P4 = 15.236;
double Still_interactions:: P4 = 14.6;
//double Still_interactions:: P4 = 40.;
//double Still_interactions:: P4 = 90;
//double Still_interactions:: P4 = 5.0;

//double Still_interactions:: P5 =  1.254;
double Still_interactions:: P5 =  1.18;

bool    Still_interactions:: on           = true;
bool    Still_interactions:: ignore_HNO   = true;
bool    Still_interactions:: FF_radii     = false;
bool    Still_interactions:: miss14       = false;
int     Still_interactions:: Update_radii = 1;

void Still_interactions::add_partial_force (Hint hint)
{
    if (hint == Whole)
    {
        add_force ();
        return;
    }

    Timer timer;
    Text  time;

    int i, count;

    if (!cache_flag_)
    {
        cache_atoms ();
        cache_bonds ();

        if (!locked())
        {
            cache_flag_ = true; //fix !!! always
            make_list   ();
            cache_flag_ = false;
        }
    }

    time += "   cache ";
    time += timer.elapsed();    timer.restart();

    try
    {
        //if (hint == Long)
            zero_forces ();      //fix remove?

        calc_nonbonded_force ();//fix 0

        time += "   nb ";
        time += timer.elapsed();    

        if (hint == Short)
        {
            timer.restart();
            //TIME

            count = atom_count ();
            for (i=0;  i<count;  ++i)
                add_atom_force (atom_[i]);
            //TIME

            time += "   atom ";
            time += timer.elapsed();    timer.restart();

            if (&bond_group_ != &nil<Bond_group>())
            {
                count = bond_count ();

                for (i=0;  i<count;  ++i)
                    add_bond_force (bond_[i]);
            }
            time += "   bond ";
            time += timer.elapsed();    
            //TIME
        }
        
        timer.restart();

        if (hint == Long)
            add_full_nonbonded_force  ();
            //TIME

        //switch (hint)
        //{
        //case Long:      add_full_nonbonded_long_force  ();    break;
        //case Short:     add_full_nonbonded_short_force ();    break;
        //default: FLAW ("No such hint.");
        //}

        time += "   add nb ";
        time += timer.elapsed();    timer.restart();
    }
    catch(...)
    {
        to_user().fatal_error ("Still_interactions: Unknown exception.");
    }

    if (!cache_flag_)
    {
        flush_force ();
//        flush_born  ();//fix down
        atom_.clear ();
        bond_.clear ();
    }

    time += "   flush ";
    time += timer.elapsed();    timer.restart();
    to_user().status (time);
}

void Still_interactions::
full_nonbonded_potential ()
{
    //TIME
    List_interactions::full_nonbonded_potential ();
    //TIME

    if (!implicit_water ())
    {
        return;
    }

    if (need_to_update_Born_radii())
    {
     //TIME
       update_Born_radii  ();
     //TIME
       nonpolar_solvation ();
     //TIME
   }

    generalized_Born ();
    //TIME
}

//void Still_interactions::
//update_Born_radii ()
//{
//    const double lIP5 = 1./P5;          //0.797448165
//    const double pi_P5 = __Pi__* P5;      
//
//    double CCF = 1.;
//
//    int i, j, count = atom_count();
//
//    for (i=0;  i<count;  ++i)
//    {
//        Atom_cache & atom1 = atom_[i];
//        //double       Gpol_i = -98.28148345;
//        //double Gpol_i = -0.5 * Constant::lI4PiEps0 / (R_vdw(atom1) + Phi + P1);
//        double Gpol_i = atom1.Gpol13___;
//        double Rvdw_i = R_vdw (atom1);
//
//        miss_neighbours (atom1);
//
//        for (j=0;  j<count;  ++j)
//        {
//            Atom_cache & atom2 = atom_[j];
//
//            double x = atom2.x___ - atom1.x___;
//            double y = atom2.y___ - atom1.y___;
//            double z = atom2.z___ - atom1.z___;
//
//            //double r_2 = x*x + y*y + z*z;
//            double r_2 = x*x + y*y + z*z + 0.000001;
//            //double r_2 = x*x + y*y + z*z + Shift;//fix
//            double r_4 = r_2 * r_2;
//            //double r_4 = r_2 * r_2 + Shift*Shift;
//                
//                double Rvdw_j = R_vdw (atom2);
//                double Dvdw = Rvdw_i + Rvdw_j;                     //3.4126
//                double factor = r_2 / (Dvdw*Dvdw);
//
//                if (factor > lIP5)
//                    CCF = 1.;
//                else
//                {
//                    CCF = .5 * (1. - cos (factor * pi_P5));
//                    CCF *= CCF;
//                }
//
//                //CCF = (factor <= lIP5) * (.5 * (1. - cos (factor * pi_P5)) - 1.) + 1.;
//                //CCF *= CCF;
//
//            double Vj = atom2.flag___.is_not (Atom_kit::missed) * atom2.V___;
//            Gpol_i += (i!=j) * P4 * Vj * CCF / r_4;
//        }
//
//        restore_neighbours ();
//
//        Gpol_i = Gpol_i < -d_mach_eps ? Gpol_i : -d_mach_eps;
//        atom1.born_radius___ = -0.5 * Constant::lI4PiEps0 / Gpol_i;
//        CHECK ("", atom1.born_radius___ > 0.);
//        //CHECK ("", atom1.born_radius___ > R_vdw (atom1) + Phi_);
//   }
//}
//
//
//
void Still_interactions::
update_Born_radii ()
{
    double lIP5 = 1./P5;          //0.797448165
    double pi_P5 = __Pi__* P5;      
    double R14_2 = R14 * R14;      

    double CCF = 1.;

    int i, j, count = atom_count();

    for (i=0;  i<count;  ++i)
        atom_data_[i].Gpol = atom_data_[i].Gpol13;

    for (i=0;  i<count;  ++i)
    {
        Atom_cache & atom1 = atom_[i];
        Atom_data  & data1 = atom_data_[i]; //x 3.8134993104836439
        double Rvdw_i = data1.Rvdw;

        miss_neighbours (atom1);

        for (j=i+1;  j<count;  ++j)
        {
            Atom_cache & atom2 = atom_[j];
            Atom_data  & data2 = atom_data_[j];

            double x = atom2.x___ - atom1.x___;
            double y = atom2.y___ - atom1.y___;
            double z = atom2.z___ - atom1.z___;

            double r_2 = x*x + y*y + z*z + Shift;//fix

            if (miss14 && atom2.flag___.is_not (Atom_kit::missed) && atom2.transparency___== 0.)
                r_2 = R14_2;

            double r_4 = r_2 * r_2;
            //double r_4 = r_2 * r_2 + Shift*Shift;
                
                double Rvdw_j = data2.Rvdw;
                double Dvdw = Rvdw_i + Rvdw_j;                     
                double factor = r_2 / (Dvdw*Dvdw);  //1.0005271666510620

                if (factor > lIP5)
                    CCF = 1.;
                else
                {
                    CCF = .5 * (1. - cos (factor * pi_P5));
                    CCF *= CCF;
                }

                //CCF = (factor <= lIP5) * (.5 * (1. - cos (factor * pi_P5)) - 1.) + 1.;
                //CCF *= CCF;

            double yes = atom2.flag___.is_not (Atom_kit::missed);
            //double yes = atom2.transparency___;

            double k = P4 * CCF / r_4;      //0.29284379976470287

            double Vj = yes * data2.V;
            data1.Gpol += k * Vj;       //-107.50654256160212

            double Vi = yes * data1.V;
            data2.Gpol += k * Vi;
        }

        restore_neighbours ();
    }

    for (i=0;  i<count;  ++i)
    {
        //Atom_cache & atom1 = atom_[i];
        Atom_data  & data1 = atom_data_[i];

        double Gpol_i = data1.Gpol;
        Gpol_i = Gpol_i < -d_mach_eps ? Gpol_i : -d_mach_eps;
        data1.born_radius = -0.5 * Constant::lI4PiEps0 / Gpol_i;
        CHECK ("", data1.born_radius > 0.);
        //CHECK ("", atom1.born_radius___ > R_vdw (atom1) + Phi_);
   }
}

void Still_interactions::
nonpolar_solvation ()
{
    double sa     = 0.;
    double k      = 4. * __Pi__ * Sigma;
    //double Rvdw   = R_vdw;
    //double Rprobe = R_probe;
    //double Rsolv  = R_solv;
    int    count  = atom_count();

    for (int i=0;  i<count;  ++i)
    {
        double ignore = 1.;
        if (ignore_HNO)
        {
            Atom_cache & atom = atom_[i];
            Element element = atom.origin___->element();
            if (element == Element (1) || 
                element == Element (7) || 
                element == Element (8))

                ignore = 0.0;
        }

        //Atom_cache & atom  = atom_[i];
        Atom_data  & data = atom_data_[i];
        //double Rvdw   = R_vdw (atom);
        double Rvdw   = data.Rvdw;
        double Rsolv  = Rvdw + R_probe;
        double ratio = Rvdw / data.born_radius;          //born radius 1.759025936434861
        sa += ignore * k * (Rsolv * Rsolv) * pow (ratio, Power);
    }
    
    surface_area_potential_ += sa;
    last_potential_ += sa;
}

void Still_interactions::
generalized_Born ()
{
    double e = 0.;
    double e_sum = 0.;
    int    count  = atom_count();
    double k = -(Constant::lI4PiEps0 / 2.) * (1. - 1. / Constant::Water_dielectric_permittivity);

    for(int j=0;  j<count;  ++j)
    {
        Atom_cache & atom_j = atom_[j];
        Atom_data  & data_j = atom_data_[j];

        double k_charge_j = k * atom_j.charge___;

        // self
        {
            double r_Born_2 = data_j.born_radius * data_j.born_radius;
            e = (k_charge_j * atom_j.charge___) / sqrt (r_Born_2);
            e_sum += e;
        }
        
        for (int i=j+1;  i<count;  ++i)
        {
            Atom_cache & atom_i = atom_[i];
            Atom_data  & data_i = atom_data_[i];

            double x = atom_j.x___ - atom_i.x___;
            double y = atom_j.y___ - atom_i.y___;
            double z = atom_j.z___ - atom_i.z___;

            double r_2 = x*x + y*y + z*z + 0.000001;
            double r_Born_2 = data_i.born_radius * data_j.born_radius;
            double D = r_2 / (4. * r_Born_2);
            double exponent = exp(-D);
            double divisor_2  = r_2 + r_Born_2 * exponent;
            double divisor  = sqrt (divisor_2);
            e = (k_charge_j * atom_i.charge___) / divisor;
            e_sum += 2 * e;
        }
    }

    generalized_Born_potential_ += e_sum;
    last_potential_ += e_sum;
}


void Still_interactions::
add_full_nonbonded_force ()
{
    //TIME
    List_interactions::add_full_nonbonded_force ();

    if (!implicit_water ())
    {
        return;
    }

    if (need_to_update_Born_radii() && (sim_counter_++ % Update_radii) == 0)
    {
    //TIME
        update_Born_radii  ();
    //TIME
        add_nonpolar_solvation_force ();
    //TIME
    }

    add_generalized_Born_force ();
    //TIME
    add_Born_radii_force  ();
    //TIME

    //add_Born_force  ();
    //TIME  
}

//void Still_interactions::
//add_full_nonbonded_long_force ()
//{
//    /// List_interactions::add_full_nonbonded_force
//    /// update_Born_radii
//
//
//    const double lIP5 = 1./P5;          //0.797448165
//    const double pi_P5 = __Pi__* P5;  
//    //const double R_short_2 = R_short_ * R_short_;
//
//    double CCF = 1.;
//
//    int i, j, count = atom_count();
//
//    for (i=0;  i<count;  ++i)
//        atom_data_[i].Gpol = atom_data_[i].Gpol13;
//
//    for (i=0;  i<count;  ++i)
//    {
//        Atom_cache & atom1 = atom_[i];
//        Atom_data  & data1 = atom_data_[i];
//        double Rvdw_i = data1.Rvdw;
//
//        //atom1.short_list___.clear ();
//        data1.short_list.clear ();
//
//        miss_neighbours (atom1);
//
//        double  atom1x = atom1.x___;
//        double  atom1y = atom1.y___;
//        double  atom1z = atom1.z___;
//
//        int count_j = i*count + count;
//
//        for (j=i*count+i+1;  j<count_j;  ++j)
//        {
//            Next_atom const & pair = pair_list_[j];
//            Atom_cache & atom2 = atom_     [pair.atom2_N___];
//            Atom_data  & data2 = atom_data_[pair.atom2_N___];
//
//            double x = atom1x - atom2.x___;
//            double y = atom1y - atom2.y___;
//            double z = atom1z - atom2.z___;
//
//            double A = pair.A___;
//            double B = pair.B___;
//
//            double r_2   = x*x + y*y + z*z + 0.000001;
//            double r_4   = r_2 * r_2;
//            double lIr   = pair.miss___ * (1.0 / sqrt (r_2));
//            double lIr_2 = lIr * lIr;    
//            double lIr_6 = lIr_2 * lIr_2 * lIr_2;
//
//            //double Rvdw_j = R_vdw (atom2);
//            double Rvdw_j = data2.Rvdw;
//            double Dvdw = Rvdw_i + Rvdw_j;          
//            double factor = r_2 / (Dvdw*Dvdw);
//
//            if (factor > lIP5)
//                CCF = 1.;
//            else
//            {
//                CCF = .5 * (1. - cos (factor * pi_P5));
//                CCF *= CCF;
//            }
//
//            int yes = atom2.flag___.is_not (Atom_kit::missed);
//            double k = P4 * CCF / r_4;
//
//            //double Vj = yes * atom2.V___;   atom1.Gpol___ += k * Vj;
//            //double Vi = yes * atom1.V___;   atom2.Gpol___ += k * Vi;
//            double Vj = yes * data2.V;   data1.Gpol += k * Vj;
//            double Vi = yes * data1.V;   data2.Gpol += k * Vi;
//
//
//            double q = Constant::lI4PiEps0 
//                * atom1.charge___ * atom2.charge___ * lIr * lIr_2;
//            double K = q;
//
//            if (true)
//            //if (r_2 > R_short_2)
//            {
//                double VdW = - (6 * (B - 2. * A * lIr_6) * lIr_6) * lIr_2;
//                K += VdW;
//            }
//            else if (atom2.transparency___)
//            {
//                //atom1.short_list___.push_back (pair);
//                data1.short_list.push_back (pair);
//            }
//
//            K *= atom2.transparency___;
//            double fx = K * x;  double fy = K * y;  double fz = K * z;
//
//            atom1.force_x___ += fx;     atom2.force_x___ -= fx;
//            atom1.force_y___ += fy;     atom2.force_y___ -= fy;
//            atom1.force_z___ += fz;     atom2.force_z___ -= fz;
//        }
//
//        restore_neighbours ();
//    }
//
//    for (i=0;  i<count;  ++i)
//    {
//        //Atom_cache & atom1 = atom_[i];
//        Atom_data  & data1 = atom_data_[i];
//
//        //double Gpol_i = atom1.Gpol___;
//        double Gpol_i = data1.Gpol;
//        Gpol_i = Gpol_i < -d_mach_eps ? Gpol_i : -d_mach_eps;
//        //atom1.born_radius___ = -0.5 * Constant::lI4PiEps0 / Gpol_i;
//        data1.born_radius = -0.5 * Constant::lI4PiEps0 / Gpol_i;
//        //CHECK ("", atom1.born_radius___ > 0.);
//        CHECK ("", data1.born_radius > 0.);
//        //CHECK ("", atom1.born_radius___ > R_vdw (atom1) + Phi_);
//    }
//
//
//    add_nonpolar_solvation_force ();
//    //TIME
//    add_generalized_Born_force ();
//    //TIME
//    add_Born_radii_force  ();
//
//}
//
//void Still_interactions::
//add_full_nonbonded_short_force ()
//{
//    int i, j, count = atom_count();
//
//    for (i=0;  i<count;  ++i)
//    {
//        Atom_cache & atom1 = atom_[i];
//        Atom_data  & data1 = atom_data_[i];
//
//        double  atom1x = atom1.x___;
//        double  atom1y = atom1.y___;
//        double  atom1z = atom1.z___;
//
//        //int count_j = atom1.short_list___.size();
//        int count_j = data1.short_list.size();
//
//        for (j=0;  j<count_j;  ++j)
//        {
//            Next_atom const & pair = data1.short_list[j];
//            Atom_cache & atom2 = atom_[pair.atom2_N___];
//
//            double x = atom1x - atom2.x___;
//            double y = atom1y - atom2.y___;
//            double z = atom1z - atom2.z___;
//
//            double A = pair.A___;
//            double B = pair.B___;
//
//            double r_2   = x*x + y*y + z*z + 0.000001;
//            double lIr_2 = 1. / r_2;    
//            double lIr_6 = lIr_2 * lIr_2 * lIr_2;
//
//            double VdW = - (6 * (B - 2. * A * lIr_6) * lIr_6) * lIr_2;
//            double K = VdW;
//
//            double fx = K * x;  double fy = K * y;  double fz = K * z;
//
//            atom1.force_x___ += fx;     atom2.force_x___ -= fx;
//            atom1.force_y___ += fy;     atom2.force_y___ -= fy;
//            atom1.force_z___ += fz;     atom2.force_z___ -= fz;
//        }
//    }
//}

void Still_interactions::
add_nonpolar_solvation_force ()
{
    double sa     = 0.;
    double k      = 4. * __Pi__ * Sigma;
    //double Rvdw   = 1.7063;
    //double Rprobe = 1.4;
    //double Rsolv  = Rvdw + Rprobe;
    int    count  = atom_count();

    for (int i=0;  i<count;  ++i)
    {
        double ignore = 1.;
        if (ignore_HNO)
        {
            Atom_cache & atom = atom_[i];
            Element element = atom.origin___->element();
            if (element == Element (1) || 
                element == Element (7) || 
                element == Element (8))

               ignore = 0.0;
        }

        //Atom_cache & atom = atom_[i];
        Atom_data & data = atom_data_[i];
        //double Rvdw   = R_vdw (atom);
        double Rvdw   = data.Rvdw;
        double Rsolv  = Rvdw + R_probe;
        double ratio = Rvdw / data.born_radius;          //born radius 1.759025936434861
        sa = ignore * k * (Rsolv * Rsolv) * pow (ratio, Power);
        data.d_born_radius = -Power * sa / data.born_radius; //-1.688
    }
}

void Still_interactions::
add_generalized_Born_force ()
{
    double e = 0.;
    int    count  = atom_count();
    double k = -(Constant::lI4PiEps0 / 2.) * (1. - 1. / Constant::Water_dielectric_permittivity);

    for (int j=0;  j<count;  ++j)
    {
        Atom_cache & atom_j = atom_[j];
        Atom_data  & data_j = atom_data_[j];

        double k_charge_j = k * atom_j.charge___;

        // self
        //{
        //    double r_Born_2 = atom_j.born_radius___ * atom_j.born_radius___;
        //    e = (k_charge_j * atom_j.charge___) / sqrt (r_Born_2);
        //    e_sum += e;
        //}

        for (int i=j;  i<count;  ++i)
        {
            Atom_cache & atom_i = atom_[i];
            Atom_data  & data_i = atom_data_[i];

            double x = atom_j.x___ - atom_i.x___;
            double y = atom_j.y___ - atom_i.y___;
            double z = atom_j.z___ - atom_i.z___;

            double r_2 = x*x + y*y + z*z + 0.000001;
            double r   = sqrt (r_2);
            //double r_Born_2 = atom_i.born_radius___ * atom_j.born_radius___;
            double r_Born_2 = data_i.born_radius * data_j.born_radius;
            double D = r_2 / (4. * r_Born_2);
            double exponent = exp(-D);
            //double divisor_2  = r_2 + r_Born_2 * exponent;
            double lIdivisor_2  = 1. / (r_2 + r_Born_2 * exponent);

            //double divisor  = sqrt (divisor_2);
            double lIdivisor  = sqrt (lIdivisor_2);


            //e = (k_charge_j * atom_i.charge___) / sqrt (r_2 + r_Born_2 * exponent);
            //e = (k_charge_j * atom_i.charge___) / divisor;
            e = (k_charge_j * atom_i.charge___) * lIdivisor;
            //e_sum += 2 * e;
            e *= 2.; 

            //double de   = -e * (r - .25 * r * exponent) / divisor_2;
            //double derb = -e * exponent * (.5 + .5 * D) / divisor_2;

            double de   = -e * (r - .25 * r * exponent) * lIdivisor_2;
            double derb = -e * exponent * (.5 + .5 * D) * lIdivisor_2;

            if (i==j)
            {
                //double d_born_radius_j = derb * atom_i.born_radius___;
                //atom_j.d_born_radius___ += d_born_radius_j;
                double d_born_radius_j = derb * data_i.born_radius;
                data_j.d_born_radius += d_born_radius_j;
            }
            else
            {
                //double d_born_radius_j = derb * atom_i.born_radius___;
                //double d_born_radius_i = derb * atom_j.born_radius___;
                //atom_j.d_born_radius___ += d_born_radius_j;
                //atom_i.d_born_radius___ += d_born_radius_i;
                double d_born_radius_j = derb * data_i.born_radius;
                double d_born_radius_i = derb * data_j.born_radius;
                data_j.d_born_radius += d_born_radius_j;
                data_i.d_born_radius += d_born_radius_i;

                de /= r;
                double K = -de;//fix

                double fx = K * x;
                double fy = K * y;
                double fz = K * z;

                atom_j.force_x___ += fx;
                atom_i.force_x___ -= fx;

                atom_j.force_y___ += fy;
                atom_i.force_y___ -= fy;

                atom_j.force_z___ += fz;
                atom_i.force_z___ -= fz;
            }
        }
    }
}


//void Still_interactions::
//add_Born_force ()
//{
//    double e = 0.;
//    int    count  = atom_count();
//    double k = -(Constant::lI4PiEps0 / 2.) * (1. - 1. / Constant::Water_dielectric_permittivity);
//    const double lIP5 = 1./P5;          //0.797448165
//    const double pi_P5 = __Pi__* P5;      
//    const double Pi4Eps0 = 1./Constant::lI4PiEps0;
//    double CCF = 1., dCCF = 0.;
//
//    for (int j=0;  j<count;  ++j)
//    {
//        Atom_cache & atom_j = atom_[j];
//
//        double k_charge_j    = k * atom_j.charge___;
//        double born_radius_j = atom_j.born_radius___;
//        double Rvdw_j        = R_vdw (atom_j);
//
//        double Gpol_j = 2. * born_radius_j * born_radius_j * Pi4Eps0;
//
//        miss_neighbours (atom_j);
//
//        for (int i=j;  i<count;  ++i)
//        {
//            Atom_cache & atom_i = atom_[i];
//
//            double x = atom_j.x___ - atom_i.x___;
//            double y = atom_j.y___ - atom_i.y___;
//            double z = atom_j.z___ - atom_i.z___;
//
//            double r_2 = x*x + y*y + z*z + 0.000001;
//            //double r_2 = x*x + y*y + z*z + Shift;
//            double r_6 = r_2 * r_2 * r_2;
//            double r   = sqrt (r_2);
//
//                double Rvdw_i = R_vdw (atom_i);
//                double Dvdw = Rvdw_i + Rvdw_j;
//                double factor = r_2 / (Dvdw*Dvdw);
//                if (factor > lIP5)
//                {
//                    CCF = 1.;
//                    dCCF = 0.;
//                }
//                else
//                {
//                    CCF = .5 * (1. - cos (factor * pi_P5));
//                    dCCF = 2. * CCF * sin (factor * pi_P5) * factor * pi_P5;
//                    CCF *= CCF;
//                }
//
//            double r_Born_2 = atom_i.born_radius___ * atom_j.born_radius___;
//            //double lIr_Born_2
//            double D = r_2 / (4. * r_Born_2);
//            double exponent = exp(-D);
//            //double divisor_2  = r_2 + r_Born_2 * exponent;
//            double lIdivisor_2  = 1. / (r_2 + r_Born_2 * exponent);
//            //double divisor  = sqrt (divisor_2);
//            double lIdivisor  = sqrt (lIdivisor_2);
//            //e = (k_charge_j * atom_i.charge___) / sqrt (r_2 + r_Born_2 * exponent);
//            //e = (k_charge_j * atom_i.charge___) / divisor;
//            e = (k_charge_j * atom_i.charge___) * lIdivisor;
//            //e_sum += 2 * e;
//            e *= 2.; 
//
//            //double de   = -e * (r - .25 * r * exponent) / divisor_2;
//            //double derb = -e * exponent * (.5 + .5 * D) / divisor_2;
//
//            double de   = -e * (r - .25 * r * exponent) * lIdivisor_2;
//            double derb = -e * exponent * (.5 + .5 * D) * lIdivisor_2;
//
//            if (i==j)
//            {
//                double d_born_radius_j = derb * atom_i.born_radius___;
//                atom_j.d_born_radius___ += d_born_radius_j;
//            }
//            else
//            {
//                double d_born_radius_j = derb * atom_i.born_radius___;
//                double d_born_radius_i = derb * atom_j.born_radius___;
//                atom_j.d_born_radius___ += d_born_radius_j;
//                atom_i.d_born_radius___ += d_born_radius_i;
//
//                de /= r;
//                double K = -de;//fix
//
//            double m = P4 * (4. * CCF - dCCF) / r_6;
//            double Vi = atom_i.flag___.is_not (Atom_kit::missed) * atom_i.V___;
//            double de_i = atom_j.d_born_radius___ * Gpol_j * Vi * m;
//
//            double born_radius_i = atom_i.born_radius___;
//            double Gpol_i = 2. * born_radius_i * born_radius_i * Pi4Eps0;
//            double Vj = atom_i.flag___.is_not (Atom_kit::missed) * atom_j.V___;
//            double de_j = atom_i.d_born_radius___ * Gpol_i * Vj * m;
//
//            K += de_i + de_j;//fix
//
//                double fx = K * x;
//                double fy = K * y;
//                double fz = K * z;
//
//                atom_j.force_x___ += fx;
//                atom_i.force_x___ -= fx;
//
//                atom_j.force_y___ += fy;
//                atom_i.force_y___ -= fy;
//
//                atom_j.force_z___ += fz;
//                atom_i.force_z___ -= fz;
//            }
//        }
//
//        restore_neighbours ();
//    }
//}
//
//
void Still_interactions::
add_Born_radii_force ()
{
    const double lIP5 = 1./P5;          //0.797448165
    const double pi_P5 = __Pi__* P5;      
    const double Pi4Eps0 = 1./Constant::lI4PiEps0;
    double CCF = 1., dCCF = 0.;
    double R14_2 = R14 * R14;      

    int count = atom_count();

    for (int j=0;  j<count;  ++j)
    {
        Atom_cache & atom_j = atom_[j];
        Atom_data  & data_j = atom_data_[j];
        //double born_radius_j = atom_j.born_radius___;
        double born_radius_j = data_j.born_radius;
        //double Rvdw_j = R_vdw (atom_j);
        double Rvdw_j = data_j.Rvdw;

        double Gpol_j = 2. * born_radius_j * born_radius_j * Pi4Eps0;

        miss_neighbours (atom_j);

        //for (int i=0;  i<count;  ++i)
        for (int i=j+1;  i<count;  ++i)
        {
            Atom_cache & atom_i = atom_[i];
            Atom_data  & data_i = atom_data_[i];

            double x = atom_j.x___ - atom_i.x___;
            double y = atom_j.y___ - atom_i.y___;
            double z = atom_j.z___ - atom_i.z___;

            double r_2 = x*x + y*y + z*z + Shift;

            if (miss14 && atom_i.flag___.is_not (Atom_kit::missed) && atom_i.transparency___== 0.)
                r_2 = R14_2;

            double r_6 = r_2 * r_2 * r_2;
            //double r_4 = r_2 * r_2;

                //double Rvdw_i = R_vdw (atom_i);
                double Rvdw_i = data_i.Rvdw;
                double Dvdw = Rvdw_i + Rvdw_j;
                double factor = r_2 / (Dvdw*Dvdw);
                if (factor > lIP5)
                {
                    CCF = 1.;
                    dCCF = 0.;
                }
                else
                {
                    CCF = .5 * (1. - cos (factor * pi_P5));
                    dCCF = 2. * CCF * sin (factor * pi_P5) * factor * pi_P5;
                    CCF *= CCF;
                }

            //double Vi = atom_i.flag___.is_not (Atom_kit::missed) * atom_i.V___;
            ////double de = atom_j.d_born_radius___ * P4 * Gpol_j * Vi * 4. / r_6;
            //double de = atom_j.d_born_radius___ * P4 * Gpol_j * Vi * (4. * CCF - dCCF) / r_6;
            ////double de = atom_j.d_born_radius___ * P4 * Gpol_j * Vi * 2. / r_4;

            double yes;
            yes = atom_i.flag___.is_not (Atom_kit::missed);

            //if (miss14)
            //    yes = atom_i.transparency___;

            if (miss14 && atom_i.transparency___ == 0)
                continue;

            double m = P4 * (4. * CCF - dCCF) / r_6;
            //double Vi = atom_i.flag___.is_not (Atom_kit::missed) * data_i.V;
            double Vi = yes * data_i.V;
            double de_i = data_j.d_born_radius * Gpol_j * Vi * m;

            double born_radius_i = data_i.born_radius;
            double Gpol_i = 2. * born_radius_i * born_radius_i * Pi4Eps0;
            //double Vj = atom_i.flag___.is_not (Atom_kit::missed) * data_j.V;
            double Vj = yes * data_j.V;
            double de_j = data_i.d_born_radius * Gpol_i * Vj * m;


            double K = de_i + de_j;//fix
            //double K = (i!=j) * de_i;//fix

            double fx = K * x;
            double fy = K * y;
            double fz = K * z;

            atom_j.force_x___ += fx;
            atom_i.force_x___ -= fx;

            atom_j.force_y___ += fy;
            atom_i.force_y___ -= fy;

            atom_j.force_z___ += fz;
            atom_i.force_z___ -= fz;
        }

        restore_neighbours ();
    }
}

void Still_interactions::
make_volumes ()
{
    sim_counter_ = 0;//fix

    assign_VdW ();
    modify_VdW ();

    int    count  = atom_count();
    double k = __Pi__* 4. / 3.;
    double m = __Pi__     / 3.;

    for (int j=0;  j<count;  ++j)
    {
        Atom_cache & atom = atom_[j];

        if (atom.origin___ == &Atom::none())
            continue;

        //double R = R_vdw (atom);
        double R = atom_data_[j].Rvdw;

        double V_sphere = k * R * R * R;
        double subV = 0.;

        for (int k=0;  k<atom.bound_count();  ++k)
        {
            Atom_cache & atom_k = atom.bound (k);

            //double R_k = R_vdw (atom_k);
            double R_k = atom_data_[atom_k.number___].Rvdw;

            //double x = atom.x___ - atom_k.x___;
            //double y = atom.y___ - atom_k.y___;
            //double z = atom.z___ - atom_k.z___;
            //double r_2 = x*x + y*y + z*z + 0.000001;
            //double r   = Mbond * sqrt (r_2);
            double K, Eq;
            force_field().bond (atom.mm_type___, atom_k.mm_type___,  K, Eq);
            Eq *= Mbond;

            //double h = R * (1. + (R_k*R_k - R*R - r*r) / (2. * R * r));
            double h = R * (1. + (R_k*R_k - R*R - Eq*Eq) / (2. * R * Eq));
            subV += m * h*h * (3.* R - h);
        }

        //atom.V___ = V_sphere - subV;
        atom_data_[j].V = V_sphere - subV;
        //CHECK ("", atom.V___ > 0);
    }

    make_G13 ();//fix
}

void Still_interactions::
make_G13 ()
{
    double K, Eq_j, Eq_k, Eq_angle, /*r_j,*/ r2, r4;
    int count = atom_count();

    for (int i=0;  i<count;  ++i)
    {
        Atom_cache & atom_1 = atom_[i];

        if (atom_1.origin___ == &Atom::none())
            continue;

        //atom_data_.back().Gpol13 = -0.5 * Constant::lI4PiEps0 / (atom_data_.back().Rvdw + Phi + P1);
        double Rvdw = atom_data_[i].Rvdw; 
        atom_data_[i].Gpol13 = -0.5 * Constant::lI4PiEps0 / (Rvdw + Phi + P1);

//continue;

        for (int j=0;  j<atom_1.bound_count();  ++j)
        {
            Atom_cache const & atom_2 = atom_1.bound (j);

            force_field().bond (atom_1.mm_type___, atom_2.mm_type___,  K, Eq_j); 
            Eq_j *= Mbond;
            r2 = Eq_j * Eq_j;     
            r4 = r2 * r2;
            //r_j = sqrt (r2);
            //atom_1.Gpol13___ += P2 * R_vdw (atom_2) / r4;
            //atom_data_.back().Gpol13 += P2 * R_vdw (atom_2) / r4;
            atom_data_[i].Gpol13 += P2 * atom_data_[atom_2.number___].V / r4;

            for (int k=0;  k<atom_2.bound_count();  ++k)
            {
                Atom_cache const & atom_3 = atom_2.bound (k);
                if (&atom_1 == &atom_3) continue;

                force_field().bond (atom_2.mm_type___, atom_3.mm_type___,  K, Eq_k);
                Eq_k *= Mbond;

                force_field().angle 
                    (atom_1.mm_type___, atom_2.mm_type___, atom_3.mm_type___, K, Eq_angle);

                r4 = r2 + Eq_k * Eq_k - 2. * Eq_j * Eq_k * cos (Eq_angle);
                r4 *= r4;
                //atom_1.Gpol13___ += P3 * R_vdw (atom_3) / r4;
                //atom_data_.back().Gpol13 += P3 * R_vdw (atom_3) / r4;
                atom_data_[i].Gpol13 += P3 * atom_data_[atom_3.number___].V / r4;
            }
        }
    }
}

void Still_interactions::
assign_VdW ()
{
    int count = atom_count();

    atom_data_.clear();
    atom_data_.reserve (count);

    for (int i=0;  i<count;  ++i)
    {
        Atom_cache & atom = atom_[i];

        if (atom.origin___ == &Atom::none())
            continue;

        atom_data_.push_back (Atom_data());
        //atom_data_.back().Rvdw   = R_vdw (atom);
        atom_data_.back().Rvdw   = atom.origin___->element() == Element (1) ? 
            RH : force_field().sigma (atom.mm_type___);
    }
}

void Still_interactions::
modify_VdW ()
{
    if (FF_radii)
        return;

    Element H("H"), N("N"), O("O"), Cl("Cl"), S("S");

    int count = atom_count();

    for (int i=0;  i<count;  ++i)
    {
        double new_Rvdw = atom_data_[i].Rvdw;
        Atom_cache & cache = atom_[i];

        if (cache.origin___ == &Atom::none())
            continue;

        Atom & atom  = *cache.origin___;
        Element element = atom.element();

        if ((element == N && (atom.bound_count() == 3 || atom.bound_count() == 2)) ||
            (element == H && (atom.bound(0).element() == O || atom.bound(0).element() == N)) ||
            (element == Cl))
            new_Rvdw *= 1.05;
        else if (element == S)
            new_Rvdw *= 1.1;

        //int element_N = atom.element().number();

        //switch (element_N)
        //{
        //case 1:
        //    if (atom.bound_count() != 0)
        //    {
        //        if      (atom.bound(0).element().number() == 7)
        //            new_Rvdw = 1.15;
        //        else if (atom.bound(0).element().number() == 8)
        //            new_Rvdw = 1.05;
        //        else
        //            new_Rvdw = 1.25;
        //    }
        //    break;

        //case 3: new_Rvdw = 1.432;    break;

        //case 6: 
        //    if      (atom.bound_count() == 3)
        //        new_Rvdw = 1.875;
        //    else if (atom.bound_count() == 2)
        //        new_Rvdw = 1.825;
        //    else
        //        new_Rvdw = 1.9;
        //    break;

        //case 7: 
        //    if      (atom.bound_count() == 4)
        //        new_Rvdw = 1.625;
        //    else if (atom.bound_count() == 1)
        //        new_Rvdw = 1.6;
        //    else
        //        new_Rvdw = 1.7063;
        //    break;

        //case 8: 
        //    if (atom.bound_count() == 1)
        //        new_Rvdw = 1.48;
        //    else
        //        new_Rvdw = 1.535;
        //    break;

        //case 9:  new_Rvdw = 1.47;    break;
        //case 10: new_Rvdw = 1.39;    break;
        //case 11: new_Rvdw = 1.992;   break;
        //case 12: new_Rvdw = 1.7;     break;
        //case 14: new_Rvdw = 1.8;     break;
        //case 15: new_Rvdw = 1.87;    break;
        //case 16: new_Rvdw = 1.775;   break;
        //case 17: new_Rvdw = 1.735;   break;
        //case 18: new_Rvdw = 1.7;     break;
        //case 19: new_Rvdw = 2.123;   break;
        //case 20: new_Rvdw = 1.817;   break;
        //case 35: new_Rvdw = 1.9;     break;
        //case 36: new_Rvdw = 1.812;   break;
        //case 37: new_Rvdw = 2.26;    break;
        //case 53: new_Rvdw = 2.1;     break;
        //case 54: new_Rvdw = 1.967;   break;
        //case 55: new_Rvdw = 2.507;   break;
        //case 56: new_Rvdw = 2.188;   break;
        //}

        atom_data_[i].Rvdw = new_Rvdw;
    }
}

}//MM
