#ifndef STILL_INTERACTIONS_H
#define STILL_INTERACTIONS_H

#ifndef LIST_INTERACTIONS_H
#include "List_interactions.h"
#endif

#ifndef LIST_INTERACTIONS_H
#include "Next_atom.h"
#endif

namespace MM
{

class Still_interactions : public List_interactions
{
    bool                implicit_water_;
    int                 sim_counter_;
    double              R_short_;       // A
   // Next_atom *         short_list_;

    struct Atom_data
    {
        double              born_radius;
        double              d_born_radius;
        double              Rvdw;
        double              V;
        double              Gpol;
        double              Gpol13;
        //double              CCF;

        Array <Next_atom>   short_list;

        Atom_data () : born_radius(0.), d_born_radius(0.), Rvdw(0.), V(0.), 
            Gpol(0.), Gpol13(0.)/*, CCF(0.)*/   {short_list.reserve (100);}
    };

    Array <Atom_data>   atom_data_;

public:
    static double       Sigma;  // non-polar solvation parameter kcal/(mol*A^2)
                                // 0.003  - J.Mol.Biol. 1998. v.284. 835-848
                                // 0.0049 - tinker

    static double       Phi;    // dielectric offset distance (A)
    static double       R_probe, Shift, Mbond, Power, P1, P2, P3, P4, P5, RH, R14;
    static bool         on, ignore_HNO, FF_radii, miss14;
    static int          Update_radii;

    enum Hint {Whole, Long, Short}/* hint_*/;

public:
    Still_interactions (Model & model) : List_interactions(model), 
        implicit_water_(true), sim_counter_(0), R_short_(5.)/*, short_list_(0)*/ {}

    virtual bool        implicit_water () const                                 {return implicit_water_ && on;}
    virtual void        set_implicit_water (bool v)                            {on = implicit_water_ = v;}

    //virtual bool        ignore_HNO () const                                     {return ignore_HNO_;}
    //virtual void        set_ignore_HNO (bool on)                                {ignore_HNO_ = on;}

   //using               List_interactions::add_force;
    virtual void        add_partial_force  (Hint hint);
    virtual void        add_long_force  () {add_partial_force (Long);}
    virtual void        add_short_force () {add_partial_force (Short);}

    virtual void        full_nonbonded_potential ();

    bool                need_to_update_Born_radii () {return true;}
    void                update_Born_radii  ();
    void                nonpolar_solvation ();
    void                generalized_Born   ();

    void                add_full_nonbonded_force ();
    //void                add_full_nonbonded_long_force ();
    //void                add_full_nonbonded_short_force ();

    void                add_nonpolar_solvation_force ();
    void                add_generalized_Born_force ();
    void                add_Born_radii_force ();

    //void                add_Born_force ();

    void                make_volumes ();
    void                make_G13 ();
    void                assign_VdW ();
    void                modify_VdW ();

protected:
    //double              R_vdw (Atom_cache const & atom) const 
    //{
    //    //if (atom.origin___->element() == Element (7))
    //    //    return 1.7063;

    //    return atom.origin___->element() == Element (1) ? 
    //        RH : force_field().sigma (atom.mm_type___);
    //}
};

}//MM

#endif //STILL_INTERACTIONS_H
