#include "Test.h"

#include "Path.h"
#include "Model_impl.h"
#include "Still_interactions.h"
#include "Lock.h"
#include "Mach_eps.h"

namespace MM
{

class Still_interactions_test : public Test
{
public:
    Still_interactions_test (const Text& suite_name) : Test (suite_name) { }

    void run()
    {
        //test_Born_radii();          count_tests();
        //test_continuum_solvation(); count_tests();
        test_NN_energy();           count_tests();
        test_CC_energy();           count_tests();
        test_NC_energy();           count_tests();

        //test_DNA();                 count_tests();
    }

    void test_Born_radii()
    {
        Model_impl model;
        model.load (Path::tests() + "NN.mlm");
        Still_interactions interactions (model);
        //Lock<Still_interactions> lock(interactions);
        interactions.cache_atoms();

        interactions.update_Born_radii ();

        //interactions.flush_born ();
        DOUBLES_EQUAL (model.atom(0).kit().born_radius(), 1.75902593, 0.001);
        //TEST ("true", model.atom(0).kit().born_radius() == 0);
    }

    void test_continuum_solvation ()
    {
        Model_impl model;
        model.load (Path::tests() + "NN.mlm");
        Still_interactions interactions (model);
        //Lock<Still_interactions> lock(interactions);
        interactions.cache_atoms();

        interactions.update_Born_radii  ();
        interactions.nonpolar_solvation ();
        interactions.generalized_Born   ();

        //interactions.flush_born ();
        DOUBLES_EQUAL (interactions.last_surface_area_potential (),     2*0.4949, 0.002);
        DOUBLES_EQUAL (interactions.last_continuum_solvation_potential (), -92.43, 0.01);
    }

    void test_NN_energy ()
    {
        Model_impl model;
        model.load (Path::tests() + "NN.mlm");
        Still_interactions interactions (model);
        interactions.set_implicit_water (true);
        double e = interactions.potential ();
        DOUBLES_EQUAL (e, -55.80, 0.2);

        interactions.add_force ();
        DOUBLES_EQUAL (model.atom(0).kit().potential_force().x(), -11.5745, 0.2);

        model.atom(0).set_x (model.atom(0).x () + d_mach_eps_2);
        double e2 = interactions.potential ();
        double d = (e - e2) / d_mach_eps_2;
        DOUBLES_EQUAL (model.atom(0).kit().potential_force().x(), d, d_mach_eps_3);
    }

    void test_CC_energy ()
    {
        Model_impl model;
        model.load (Path::tests() + "CC.mlm");
        Still_interactions interactions (model);
        interactions.set_implicit_water (true);
        double e = interactions.potential ();
        DOUBLES_EQUAL (e, 13.2914, 0.2);

        interactions.add_force ();
        DOUBLES_EQUAL (model.atom(0).kit().potential_force().x(), -55.1780, 0.2);

        model.atom(0).set_x (model.atom(0).x () + d_mach_eps_2);
        double e2 = interactions.potential ();
        double d = (e - e2) / d_mach_eps_2;
        DOUBLES_EQUAL (model.atom(0).kit().potential_force().x(), d, d_mach_eps_3);
    }

    void test_NC_energy ()
    {
        Model_impl model;
        model.load (Path::tests() + "NC.mlm");
        Still_interactions interactions (model);
        interactions.set_implicit_water (true);
        double e = interactions.potential ();
        DOUBLES_EQUAL (e, -25.5505, 0.2);

        interactions.add_force ();
        DOUBLES_EQUAL (model.atom(0).kit().potential_force().x(), -17.9325, 0.2);

        model.atom(0).set_x (model.atom(0).x () + d_mach_eps_2);
        double e2 = interactions.potential ();
        double d = (e - e2) / d_mach_eps_2;
        DOUBLES_EQUAL (model.atom(0).kit().potential_force().x(), d, d_mach_eps_3);
    }

    void test_DNA ()
    {
        Model_impl model;
        model.load (Path::tests() + "B_DNA_6.hin");
        Still_interactions interactions (model);
        interactions.set_implicit_water (true);
        double e = interactions.potential ();
        DOUBLES_EQUAL (e, 2520.61, 0.01);

        interactions.add_force ();
        DOUBLES_EQUAL (model.atom(0).kit().potential_force().x(), -21.4093, 0.0001);

        model.atom(0).set_x (model.atom(0).x () + d_mach_eps_2);
        double e2 = interactions.potential ();
        double d = (e - e2) / d_mach_eps_2;
        DOUBLES_EQUAL (model.atom(0).kit().potential_force().x(), d, d_mach_eps_3);
    }
};

Still_interactions_test test_still_interactions ("demo"); //correctness

}//MM
