#include "Yakub_Ronchi_interactions.h"

#include "Model_kit.h"
#include "anint.h"

#include <math.h>

namespace MM
{

//Yakub_Ronchi_interactions::
//Yakub_Ronchi_interactions (Atom_group & atom_group, Force_field & ff)
//:
//    Conventional_interactions (atom_group, ff)
//{
//}
//
//Yakub_Ronchi_interactions::
//Yakub_Ronchi_interactions (Atom_group & atom_group,
//    Bond_group & bond_group, Force_field & ff)
//:
//    Conventional_interactions (atom_group, bond_group, ff)
//{
//}

Yakub_Ronchi_interactions::
Yakub_Ronchi_interactions (Model & model/*, Force_field & ff*/)
:
    Conventional_interactions (model/*, ff*/)
{
}

void Yakub_Ronchi_interactions::
nonbonded_potential (Atom_cache & atom)
{
    double Rm      = force_field().Rm_;
    
    double  L = 0., lIL = 0., LI2 = 0.;
    bool    box = false;
    if (model_.kit().boundary_conditions ().type_name() == "Box")
    {
        box = true;
        L = model_.kit().boundary_conditions ().box ().a_mod_rectangle();
        lIL = 1. / L;
        LI2 = L / 2.;

        if (Rm > LI2)
            Rm = LI2;
    }

    const double Rm_2    = Rm * Rm;
    const double lIRm    = 1. / Rm;
    //const double lIRm_3  = 1. / (Rm*Rm*Rm);
    //const double lI2Rm_3 = 0.5 * lIRm_3;

    //double  lIepsilon = 1. / epsilon_;//fix
    double  lIepsilon = 1. / force_field().eps;//fix

    double  A = 0., B = 0.;

    int     atom_number = atom.number___;
    int     count       = atom_count();

    double  charge_332  = atom.charge___ * EPS0; //331.8;
    //double  e           = 0;
    double  P_VdW = 0., P_el = 0., P_hph = 0., Yakub_Ronchi = 0.;
    double  x, y, z;
    double  r_2, lIr, lIr_6, lIr_eps;

    Force_field::VdWaals * vdw = force_field().VdW_par [atom.mm_type___];

    double  ax = atom.x___;
    double  ay = atom.y___;
    double  az = atom.z___;

    miss_neighbours (atom);

    // 0.00315
    for (int i=atom_number+1;  i<count;  ++i)
    //for (int i=atom_number+1;  i<count;  i+=4)
    {
        Atom_cache & current_atom  = atom_[i];

        Force_field::VdWaals & tmp = vdw [current_atom.mm_type___];  //!!!
        A = tmp.A;
        B = tmp.B; 

        x = ax - current_atom.x___;
        y = ay - current_atom.y___;
        z = az - current_atom.z___;

        if (box)
        {
            //x = fmod (x, L);
            //y = fmod (y, L);
            //z = fmod (z, L);
            x -= L * anint (x * lIL);
            y -= L * anint (y * lIL);
            z -= L * anint (z * lIL);
        }

        r_2     = x*x + y*y + z*z;
        lIr     = 1.0 / sqrt (r_2);
        lIr_eps = lIr * lIepsilon;
        lIr_6   = lIr * lIr;    lIr_6 *= lIr_6 * lIr_6;

        //Yakub_Ronchi = 1. + (r_2 * r_2 * lIr) * lI2Rm_3;
        double rIRm = r_2 * lIr * lIRm;
        Yakub_Ronchi   = 1. + .5 * rIRm * (rIRm * rIRm - 3.); 

        P_VdW += current_atom.transparency___ * 
            (A * lIr_6 - B ) * lIr_6;

        P_el += current_atom.transparency___ * (r_2 < Rm_2) *
            current_atom.charge___ * charge_332 * lIr_eps * Yakub_Ronchi;
    }

    process_14 (atom, Potential);
    restore_neighbours ();

    Van_der_Waals_potential_ += P_VdW;
    electrostatic_potential_ += P_el;
    hydrophobic_potential_   += P_hph;

    last_potential_ += P_el + P_VdW + P_hph;
}

void Yakub_Ronchi_interactions::
add_nonbonded_force (Atom_cache & atom)
{
    double Rm      = force_field().Rm_;

    double  L = 0., lIL = 0., LI2 = 0.; 
    bool    box = false;
    if (model_.kit().boundary_conditions ().type_name() == "Box")
    {
        box = true;
        L = model_.kit().boundary_conditions ().box ().a_mod_rectangle();
        lIL = 1. / L;
        LI2 = L / 2.;

        if (Rm > LI2)
            Rm = LI2;
    }
    
    const double Rm_2    = Rm * Rm;
    //const double lIRm    = 1. / Rm;
    const double lIRm_3  = 1. / (Rm*Rm*Rm);
    //const double lI2Rm_3 = 0.5 * lIRm_3;

    //double  lIepsilon = 1. / epsilon_;
    double  lIepsilon = 1. / force_field().eps;//fix
    double  A = 0., B = 0.;

    int     atom_number = atom.number___;
    int     count       = atom_count();

    double  charge_332  = atom.charge___ * EPS0; //331.8;   
    //double  charge_332  = atom.charge___ * EPS0 / 2.; //331.8;   

    double  x, y, z;
    double  r_2, lIr, lIr_2, lIr_6/*, lIr_eps*/;

    Force_field::VdWaals * vdw = force_field().VdW_par [atom.mm_type___];

    double  ax = atom.x___;
    double  ay = atom.y___;
    double  az = atom.z___;

    miss_neighbours (atom);

    for (int i=atom_number+1;  i<count;  ++i)
    {
        Atom_cache & current_atom = atom_[i];

        Force_field::VdWaals & tmp = vdw [current_atom.mm_type___];
        A = tmp.A;
        B = tmp.B;    

        x = ax - current_atom.x___;
        y = ay - current_atom.y___;
        z = az - current_atom.z___;

        if (box)
        {
            x -= L * anint (x * lIL);
            y -= L * anint (y * lIL);
            z -= L * anint (z * lIL);
        }

        r_2     = x*x + y*y + z*z;
        lIr     = 1.0 / sqrt (r_2);
        lIr_2   = lIr * lIr;
        //lIr_eps = lIr * lIepsilon;
        lIr_6 = lIr_2 * lIr_2 * lIr_2;

        double Kf_Yakub_Ronchi      = lIRm_3     - lIr_2 * lIr; 

        double F_VdW = - (6 * (B - 2 * A * lIr_6) * lIr_6) * lIr_2;

        double F_el = - lIepsilon * current_atom.charge___ * charge_332 
            * Kf_Yakub_Ronchi * (r_2 < Rm_2);

        double K = current_atom.transparency___ * (F_el + F_VdW);//fix P

        double fx = K * x;
        double fy = K * y;
        double fz = K * z;

        atom.        force_x___ += fx;
        current_atom.force_x___ -= fx;

        atom.        force_y___ += fy;
        current_atom.force_y___ -= fy;

        atom.        force_z___ += fz;
        current_atom.force_z___ -= fz;
    }
    process_14 (atom, Force);
    restore_neighbours ();
}

}//MM
