#include "Collect_temperature.h"

#include "Atom.h"
#include "Atom_kit.h"
#include "Atom_group.h"
#include "System_of_Units.h"
#include "Path.h"

namespace MM
{

Collect_temperature::
Collect_temperature (Atom_group & atom_group)
:
    Collect_scalar (atom_group)
{
//    new_graph_presentation ("temperature");
    new_file_presentation (Path::report() +"dyn_Temperature.xls");
}

Monitor * Collect_temperature::
clone (Atom_group & atom_group, Text const & /*args*/) const
{
    return new Collect_temperature (atom_group);
}

double Collect_temperature::
extract_current_value ()
{
    int     count = atom_group().atom_count();
    double  sum = 0.;

    for (int i=0;  i<count;  ++i)
    {
        Atom & atom = atom_group().atom (i);
        double mass = atom.mass();
        double velocity_2 = atom.kit().velocity().length_2();
        sum += mass * velocity_2;
    }

    //0.001987191 kcal mol-1 K-1 * 4.184  = 0.008 314 51 kJ mol-1 K-1
    double kB = System_of_Units::singleton().Boltzmann_constant();

    return  sum / (3. * count * kB);
}

}//MM
