#ifndef ANDERSEN_THERMOSTAT_H
#define ANDERSEN_THERMOSTAT_H

#ifndef THERMOSTAT_H
#include "Thermostat.h"
#endif

#ifndef VECTOR_3D_IMPL_H
#include "Vector_3D_impl.h"
#endif

#ifndef HYDROGEN_ATOMS_H
#include "Hydrogen_atoms.h"
#endif

#include "__rand.h"

#include "Flaw.h"

namespace MM
{

class Andersen_thermostat : public Thermostat
{
    double          nu_;
    T_Base_random   rand_;

public:
    Andersen_thermostat (Atom_group & atom_group, MD & simulation)
    :   
        Thermostat (atom_group, simulation), 
            nu_(0.1), 
            //nu_(0.25), 
            rand_(0., 1.)
    {
        rand_.set_method (T_Base_random::Mix);
    }

    char const* title () const      {return "Andersen";}

    void            set_nu (double nu) {nu_ = nu;}

    virtual void    handle_execute ()
    {
        Hydrogen_atoms group (atom_group_);
        //Atom_group & group = atom_group_;

        double dt    = MD_simulation_.step_length ();                           // time step of integration algorithm
        double T0    = T0_ * System_of_Units::singleton().Boltzmann_constant();
        int    count = group.atom_count();

        for (int i=0;  i<count;  ++i)
        {
            if (rand_.next() < nu_*dt)
            {
                Atom & atom  = group.atom (i);
                double mass  = atom.mass();
                double sigma = sqrt (T0 / mass);
                double x     = sigma * gauss ();
                double y     = sigma * gauss ();
                double z     = sigma * gauss ();
                atom.kit().set_velocity (Vector_3D_impl (x, y, z));
            }
        }
    }

    double gauss ()
    {
        double R_2 = 2., v1, v2;

        while (R_2 >= 1.)
        {
            v1 = 2. * rand_.next() - 1.;
            v2 = 2. * rand_.next() - 1.;
            R_2 = v1*v1 + v2*v2;
        }

        double result = v1 * sqrt (-2. * log(R_2) / R_2);
        return result;
    }
};

}//MM

#endif //ANDERSEN_THERMOSTAT_H
