#include "MD_method.h"

#include "Position_leapfrog.h"
#include "Velocity_leapfrog.h"
#include "Position_verlet.h"
#include "Velocity_verlet.h"
#include "Verlet.h"
#include "MTS_MD.h"
#include "OPVL.h"
#include "OVVL.h"
#include "VEFRL.h"
#include "PEFRL.h"

#include "Create.h"

namespace MM
{
Position_leapfrog * prototype_of_Position_leapfrog  = new Position_leapfrog (Prototype());
Velocity_leapfrog * prototype_of_Velocity_leapfrog  = new Velocity_leapfrog (Prototype());
Position_verlet *   prototype_of_Position_verlet    = new Position_verlet   (Prototype());
Velocity_verlet *   prototype_of_Velocity_verlet    = new Velocity_verlet   (Prototype());
Verlet *            prototype_of_Verlet             = new Verlet            (Prototype());
MTS_MD *            prototype_of_MTS_MD             = new MTS_MD            (Prototype());
OPVL *              prototype_of_OPVL               = new OPVL              (Prototype());
OVVL *              prototype_of_OVVL               = new OVVL              (Prototype());
VEFRL *             prototype_of_VEFRL              = new VEFRL             (Prototype());
PEFRL *             prototype_of_PEFRL              = new PEFRL             (Prototype());

// ======================================================
MD_method * MD_method::
create (Text const & method)
{
    for (int i=0;  i<prototype_count();  ++i)
        if (method == prototype (i). class_name ())
            return prototype (i).clone ();

    FLAW (method + " - no such algorithm.");
}

// fix to template ======================================
namespace
{
    Array_of_own <MD_method> & prototype_list ()
    {
        static Array_of_own <MD_method>   instance;
        return instance;
    }
}

int MD_method::
prototype_count ()
{
    return prototype_list().size();
}

MD_method & MD_method::
prototype (int n)
{
    return prototype_list()[n];
}

void MD_method::
adopt_prototype (MD_method * method)
{
    prototype_list().push_back (method);
}
//==================================================================

MD_method::
~MD_method ()
{
    simulation_ = 0;
}

void MD_method::
start ()
{
    simulation().algorithm().description
        (Text("start  ") + simulation ().title());
}

void MD_method::
finish ()
{
    simulation().algorithm().description
        (Text("finish ") + simulation ().title());
}

void MD_method::
F (MD_subject::Hint hint)
{
    INVARIANT;
    simulation_->interface_for(this).F (hint);
}
void MD_method::
kick (double h)
{
    INVARIANT;
    simulation_->interface_for(this).kick (h);
}

void MD_method::
drift(double h)
{
    INVARIANT;
    simulation_->interface_for(this).drift(h);
}

}//MM
