#include "Model_as_md_subject.h"

#include "Program_title.h"
#include "Model.h"
#include "Model_kit.h"
#include "Atom.h"
#include "Atom_kit.h"
#include "Atom_group.h"
#include "Interaction.h"
#include "Mass.h"
#include "Vector_3D_impl.h"

#include "Log.h"

namespace MM
{

Model_as_md_subject::
Model_as_md_subject (Model & model)
:
    model_(model),// interaction_(interaction),
    masses_(new Mass (model)),
    //atom_group_as_md_subject_(model_, interaction_, new Mass(model))
    first_bad_iteration_  (true),
#ifndef ASCALAPH
    lIforce_limit_  (1. / 1000.)
#else
    lIforce_limit_  (1. / 300.)
    //lIforce_limit_  (1. / 100)
#endif
{
}

void Model_as_md_subject::
F (Hint hint)
{
    update_interactions (hint);
}

void Model_as_md_subject::
kick (double dT)
{
    //update_interactions (hint);
    int size = model_.atom_count();

    for (int i=0;  i<size;  ++i)
    {
        // v = v0 + dT * F / m
        Atom &          atom = model_.atom(i);

        if (atom.kit().was_frozen())
            continue;

        Vector_3D_impl  dV (atom.kit().potential_force());
        dV *= dT * masses_().inverse(i);

        atom.kit().add_velocity (dV);
    }
}

void Model_as_md_subject::
drift (double dT)
{
    int size = model_.atom_count();

    for (int i=0;  i<size;  ++i)
    {
        // x = x0 + dT * v0
        Atom &          atom = model_.atom(i);

        if (atom.kit().was_frozen())
            continue;

        Vector_3D_impl  movement (atom.kit().velocity());
        movement *= dT;
        atom.move (movement, true);
    }
    //update_interactions ();
}

void Model_as_md_subject::
update ()
{
    int size = model_.atom_count();

    for (int i=0;  i<size;  ++i)
    {
        Atom & atom = model_.atom(i);
        atom.update ();
    }

    model_.kit().event().conformation_changed ();
}

void Model_as_md_subject::
update_interactions (Hint hint)
{
    //fix describe
    //fix dynamic check or DbC
    int size = model_.atom_count();

    for (int i=0;  i<size;  ++i)
        model_.atom(i).kit().set_potential_force (Vector_3D_impl (0,0,0));

    bool ad_hoc = model_.kit().interaction().common().ad_hoc();
    model_.kit().force_field().start_work (!ad_hoc);
    {
        switch (hint)
        {
        case Whole:
            model_.kit().interaction().add_force ();
            break;
        case Long:
            model_.kit().interaction().add_long_force ();
            break;
        case Short:
            model_.kit().interaction().add_short_force ();
            break;
        }
    }
    if  (!ad_hoc && model_.kit().force_field().was_problems ())
        throw Text ("Force Field problem");

    cut_large_forces ();
}

void Model_as_md_subject::
cut_large_forces ()
{
    bool log_it = first_bad_iteration_;
    int size = model_.atom_count();

    for (int i=0;  i<size;  ++i)
    {
        Atom &         atom  = model_.atom (i);
        Vector_3D_impl force (atom.kit().potential_force ());

        //double fac = 0.0001 * dT * dT;
//        double factor = force.length() * limit * masses_().inverse(i);
        double factor = force.length() * lIforce_limit_;

        if (factor > 1.)
        {
            force *= 0.1 / factor;
            atom.kit().set_potential_force (force);
             
            Vector_3D_impl v (atom.kit().velocity ());
            v *= 0.9;                                       //fix
            atom.kit().set_velocity (Vector_3D_impl(v));

            if (log_it)
            {
                Text message ("!!! Force on atom ");
                message += i;
                message += " ";
                message += atom.kit(). mm_type_c_str();
                message += " scaled by ";
                message += 1. / factor;
                message += "\n";
                log () << message;

                first_bad_iteration_ = false;
            }
        }
    }
}

}//MM


