#ifndef THERMOSTAT_H
#define THERMOSTAT_H

namespace MM
{
class Temperature_unit;
class Atom_group;
class MD;

class Thermostat
{
protected:
    Atom_group &                atom_group_;
    MD         &                MD_simulation_;
    double                      T0_;                // desired temperature
    bool                        thermostat_flag_;

public:
    explicit        Thermostat (Atom_group & atom_group, MD & simulation);

    virtual        ~Thermostat () { }

    static Thermostat & none ();

    bool            is_on  () const                                             {return  thermostat_flag_;}
    bool            is_off () const                                             {return !thermostat_flag_;}
    void            turn (bool on)                                              {thermostat_flag_ = on;}

    virtual char const* title () const =0;

    virtual void    execute ()                                                  {if (is_on()) handle_execute ();}
    virtual void    handle_execute () =0;

    double              calc_temperature () const;
    double              calc_temperature (Atom_group const & ) const;
    
    Temperature_unit    desired_temperature () const;
    void                set_desired_temperature (Temperature_unit const & t);
};

}//MM

#endif //THERMOSTAT_H
