*
*   Molecular Dynamics Expanded Ensemble
*
*   Chamge of ensemble
*
*======================= CHENS ================================
*
	subroutine CHENS
	include "mdee.h" 
	real*4 rand 
	external rand
C  Decide up or down
	if(TASKID.eq.MAST)then
	  if(rand().lt.0.5)then
	    ME1=ME-1
	  else
	    ME1=ME+1
	  end if
	end if
C---> report result (ME1) to other nodes
C  Increase counter of ensemble change attempts
	ICHE=ICHE+1
	if(ME1.lt.1.or.ME1.gt.NE)return
	time20=cputime(0.d0)   
C   Compute change of energy due to reciprocal part of Ewald  
	call DIFFUR(ME1,EFN,EFO)
C   Compute change of energy due to other interactions
	call DIFEN(ME1,EEN,EEO)

C  Here should be: send total energy charge to "master" node

	DEF=EFN-EFO
	DEE=EEN-EEO  
	DE=(DEE+DEF)*BETA+EE(ME1)-EE(ME)
	ENNN=(EEN+EFN)*PERMOL
	ENNO=(EEO+EFO)*PERMOL
	if(IPRINT.ge.6)write(*,'(a5,I5,a4,I5,a4,f12.4,3x,3f11.3)')
     + 'from ',ME,' to ',ME1,' DE=',DE,ENNN,ENNO,EE(ME1)-EE(ME)
C   MC transition - Metropolis rule
	if(DE.gt.0)then
	  if(DE.gt.20)go to 100
	  if(exp(-DE).lt.rand())go to 100
	end if
C  Successful transition
	ME=ME1
	ICHU=ICHU+1
	TRPRES=TPRES+(1.d0-EC(ME))*NDEL*BOLTZ*TEMP*1.d-5/(VOL*UNITL**3)
C   Redefine interactions
	call RECINT
C  Add to transition matrix
      do I=1,ME-1
        if(IWALK(I,ME).le.IWALK(ME,I))
     +  IWALK(I,ME)=IWALK(I,ME)+1
      end do
      do I=ME+1,NE
        if(IWALK(I,ME).lt.IWALK(ME,I))
     +  IWALK(I,ME)=IWALK(I,ME)+1
      end do
 100	time2=time2+cputime(time20)
	return
	end                     
*	
*======================= RECINT ===============================
*
      subroutine RECINT
      include "mdee.h"
      NDEL=0
      do ITYP=1,NTYPES
	if(IEE(ITYP).eq.0)then
	  ESCI=EC(ME)
	  NDEL=NDEL+NSPEC(ITYP) 
	else
	  ESCI=1.d0
	end if   
        DO I       =  ISADR(ITYP)+1,ISADR(ITYP+1)
	  do JTYP=1,NTYPES
	    if(IEE(JTYP).eq.0)then
	      ESCJ=EC(ME) 
	    else
	      ESCJ=1.d0
	    end if   
	    ESCL=ESCI*ESCJ
            DO J       =  ISADR(JTYP)+1,ISADR(JTYP+1)
              EPSI       =   ESCL*EPSIL(I)*EPSIL(J)
              SIGM       =   SIGMA(I)+SIGMA(J)
              TSIX       = -EPSI*SIGM**6
              TTWL       = -TSIX*SIGM**6
              ONE(I,J)   =  ESCL*CHARGE(I)*CHARGE(J)*COULF
              SIX(I,J)   =  TSIX
              TWL(I,J)   =  TTWL
            END DO! OF J
          END DO
	end do
      end do
      return
      end
*
*===================== DIFFUR ===========================================
*
C
C  Compute difference of energy doe to reciprocal part of the Ewald sum
C
      subroutine DIFFUR(ME1,EFN,EFO)
      include "mdee.h"
      DATA TWOPI/6.2831853072D0/
*
      timef0	= cputime(0.d0)
*
      QSS         = 0.D0
      CL	= BOXL
      CFUR=4.d0*PI*COULF/VOL     
      DO ITYP     = 1,NTYPES
        QSS         = QSS+QSUM(ITYP)
      END DO! OF ITYP
      IF(QSS.EQ.0.D0) RETURN
*
      EFN     = 0.D0
      EFO	= 0.d0 
      ESFF=0.d0
*   reciprocal space Evald
      IBEG = NUMTASK-TASKID 
      do IKV=IBEG,NKV,NUMTASK
        SCSO=0.
        SSNO=0.
	SCSN=0.
	SSNN=0.
 	RKX=RX(IKV)
	RKY=RY(IKV)
	RKZ=RZ(IKV)
C$DIR NO_RECURRENCE
        do I=1,NSTOT
	  IC=NSITE(I)
	  ITYP=ITYPE(I)
	  SCP=RKX*SX(I)+RKY*SY(I)+RKZ*SZ(I)
          SSC=sin(SCP)
	  CSC=cos(SCP)
	  if(IEE(ITYP).eq.0)then
	    QO=CHARGE(IC)*EC(ME)
	    QN=CHARGE(IC)*EC(ME1) 
	  else
	    QO=CHARGE(IC)
	    QN=QO
	  end if
          SSNO=SSNO+QO*SSC
          SCSO=SCSO+QO*CSC
          SSNN=SSNN+QN*SSC
          SCSN=SCSN+QN*CSC
        end do
	AKC = CFUR*AK(IKV)
        EFO=EFO+(SSNO**2+SCSO**2)*AKC
        EFN=EFN+(SSNN**2+SCSN**2)*AKC
      end do
*
*  contribution from ETERMS excluded
*	FACTR=  0.001*ENERF/FNOP
*	if(IPRINT.ge.7)write(*,'(a,4f12.5)')' FURtot:',FACTR*EFN,FACTR*EFO
      return
      end  
*
*====================== DIFEN ========================================
*                                                                     
	subroutine DIFEN(ME1,EEN,EEO)
	include "mdee.h"
	EEN=0.d0
	EEO=0.d0         
	ELJO=0.
	ESTO=0.
	ELJN=0.
	ESTN=0.
	DO INP      = 1,MBLN
	  ISP=iabs(NBL1(INP))
	  JSP=NBL2(INP)
	  ITYP=ITYPE(ISP)
	  JTYP=ITYPE(JSP)
	  if(IEE(ITYP)*IEE(JTYP).ne.1)then
	  ISB =NSITE(ISP)
  	  JSB =NSITE(JSP)
	  IMOL=NNUM(ISP)
	  JMOL=NNUM(JSP)
	  if(IMOL.ne.JMOL)then
  	  A6O =SIX(ISB,JSB)
	  B12O=TWL(ISB,JSB)
	  QIJO=ONE(ISB,JSB)
	  if(IEE(ITYP).eq.0)then
	    ESCI=EC(ME1) 
	  else
	    ESCI=1.d0
	  end if
	  if(IEE(JTYP).eq.0)then
	    ESCJ=EC(ME1) 
	  else
	    ESCJ=1.d0
	  end if
	  ESCL=ESCI*ESCJ
	  EPSI	= ESCL*EPSIL(ISB)*EPSIL(JSB)
	  SS6=(SIGMA(ISB)+SIGMA(JSB))**6
        A6N        = -EPSI*SS6
        B12N       = -A6N*SS6
        QIJN       =  ESCL*CHARGE(ISB)*CHARGE(JSB)*COULF
        DX           = SX(ISP)-SX(JSP)
        DY           = SY(ISP)-SY(JSP)
        DZ           = SZ(ISP)-SZ(JSP)
*                  call PBC(DX,DY,DZ)
        if(DX.gt. HBOXL)DX=DX-BOXL
        if(DX.lt.-HBOXL)DX=DX+BOXL
        if(DY.gt. HBOYL)DY=DY-BOYL
        if(DY.lt.-HBOYL)DY=DY+BOYL
        if(DZ.gt. HBOZL)DZ=DZ-BOZL
        if(DZ.lt.-HBOZL)DZ=DZ+BOZL
*  
	RR           = DX**2+DY**2+DZ**2
	R2I	=  1.d0/RR
        R1           = DSQRT(RR)
*  Electrostatic interaction
        ALPHAR       = ALPHAD*R1
	  TTT          = 1.D0/(1.D0+B1*ALPHAR)
        EXP2A        = DEXP(-ALPHAR**2)
        ERFCR = ((((A5*TTT+A4)*TTT+A3)*TTT+A2)*TTT+A1)*TTT*EXP2A/R1
        EESO          = QIJO*ERFCR
        EESN          = QIJN*ERFCR
*  LJ interaction
        R6I          = R2I*R2I*R2I
        ESRO          = (     A6O+      B12O*R6I)*R6I 
        ESRN          = (     A6N+      B12N*R6I)*R6I
        EEN=EEN+EESN+ESRN
	EEO=EEO+EESO+ESRO
	ELJN=ELJN+ESRN
	ESTN=ESTN+EESN
	ELJO=ELJO+ESRO
	ESTO=ESTO+EESO
	end if !  IMOL.ne.JMOL
*	if(IPRINT.ge.7.and.ESCL.ne.1.d0)write(*,'(2i5,2f12.5,2e14.6)')
*     +ISP,JSP,ENERF*0.001*(ESRO+EESO),ENERF*0.001*(ESRN+EESN),A6O,QIJO
*
	end if
      END DO! OF INP					
*    Local forces 
	DO INP      = 1,MBSH
	  ISP=iabs(NBS1(INP))
	  JSP=NBS2(INP)
	  ITYP=ITYPE(ISP)
	  JTYP=ITYPE(JSP)
	  if(IEE(ITYP)*IEE(JTYP).ne.1)then
	  IMOL=NNUM(ISP)
	  JMOL=NNUM(JSP)
	  if(IMOL.ne.JMOL)then
	  ISB =NSITE(ISP)
  	  JSB =NSITE(JSP)
  	  A6O =SIX(ISB,JSB)
	  B12O=TWL(ISB,JSB)
	  QIJO=ONE(ISB,JSB)
	  if(IEE(ITYP).eq.0)then
	    ESCI=EC(ME1) 
	  else
	    ESCI=1.d0
	  end if
	  if(IEE(JTYP).eq.0)then
	    ESCJ=EC(ME1) 
	  else
	    ESCJ=1.d0
	  end if
	  ESCL=ESCI*ESCJ
	  EPSI	= ESCL*EPSIL(ISB)*EPSIL(JSB)
	  SS6=(SIGMA(ISB)+SIGMA(JSB))**6
        A6N        = -EPSI*SS6
        B12N       = -A6N*SS6
        QIJN       =  ESCL*CHARGE(ISB)*CHARGE(JSB)*COULF
        DX           = SX(ISP)-SX(JSP)
        DY           = SY(ISP)-SY(JSP)
        DZ           = SZ(ISP)-SZ(JSP)
*                  call PBC(DX,DY,DZ)
        if(DX.gt. HBOXL)DX=DX-BOXL
        if(DX.lt.-HBOXL)DX=DX+BOXL
        if(DY.gt. HBOYL)DY=DY-BOYL
        if(DY.lt.-HBOYL)DY=DY+BOYL
        if(DZ.gt. HBOZL)DZ=DZ-BOZL
        if(DZ.lt.-HBOZL)DZ=DZ+BOZL
*  
	  RR           = DX**2+DY**2+DZ**2
	  R2I	= 1.d0/RR
        R1           = DSQRT(RR)
*  Electrostatic interaction
        ALPHAR       = ALPHAD*R1
	  TTT          = 1.D0/(1.D0+B1*ALPHAR)
        EXP2A        = DEXP(-ALPHAR**2)
        ERFCR = ((((A5*TTT+A4)*TTT+A3)*TTT+A2)*TTT+A1)*TTT*EXP2A/R1
        EESO          = QIJO*ERFCR
        EESN          = QIJN*ERFCR
*  LJ interaction
        R6I          = R2I*R2I*R2I
        ESRO          = (     A6O+      B12O*R6I)*R6I 
        ESRN          = (     A6N+      B12N*R6I)*R6I
        EEN=EEN+EESN+ESRN
	EEO=EEO+EESO+ESRO 
	ELJN=ELJN+ESRN
	ESTN=ESTN+EESN
	ELJO=ELJO+ESRO
	ESTO=ESTO+EESO
	end if   ! of IMOL.ne.JMOL
*
	end if
      END DO! OF INP	     
*  Long range corrections
      RC          = DSQRT(RSSQ)
      FAC  = 4.*PI/VOL 
      ELRN = 0.
      ELRO = 0.
      DO ITYP = 1,NTYPES
        ISBEG       = ISADR(ITYP)+1
        ISEND       = ISADR(ITYP +1)
        NSPI        = NSPEC(ITYP)
	if(IEE(ITYP).eq.0)then
	  ESCIN=EC(ME1)
	  ESCIO=EC(ME)
	else
	  ESCIN=1.d0
	  ESCIO=1.d0
	end if
        DO JTYP = ITYP ,NTYPES
          MT          = MDX  (ITYP,JTYP)
          JSBEG       = ISADR(JTYP)+1
          JSEND       = ISADR(JTYP +1)
          NSPJ        = NSPEC(JTYP)
          FNOPIJ      = DFLOAT(NSPI*NSPJ)
          if(ITYP.eq.JTYP)FNOPIJ=0.5*FNOPIJ
	  if(IEE(JTYP).eq.0)then
	    ESCJN=EC(ME1)
	    ESCJO=EC(ME)
	  else
	    ESCJN=1.d0
	    ESCJO=1.d0
	  end if
          ESCN=ESCIN*ESCJN
	  ESCO=ESCIO*ESCJO
          DO IS = ISBEG,ISEND
            DO JS = JSBEG,JSEND
	      EP          = EPS(IS,JS)
              SI          = SIG(IS,JS)  
              ELR         = EP*SI**3*FAC*
     X(( 1.D0/9.D0)*(SI/RC)**9 - ( 1.D0/3.D0)*(SI/RC)**3)*FNOPIJ
	      ELRN = ELRN + ESCN*ELR
	      ELRO = ELRO + ESCO*ELR
            end do
	  end do
        end do
      end do
      EEN = EEN + ELRN
      EEO = EEO + ELRO
      if(IPRINT.ge.7)write(*,'(a,2f11.5,a,2f12.5,a,2f11.5)')
     +'LJ:',PERMOL*ELJN,PERMOL*ELJO,' El:',PERMOL*ESTN,
     +PERMOL*ESTO,' LR:',PERMOL*ELRN,PERMOL*ELRO
      return
      end 
*
